-- This Source Code Form is subject to the terms of the bCDDL, v. 1.1.
-- If a copy of the bCDDL was not distributed with this
-- file, You can obtain one at http://beamng.com/bCDDL-1.1.txt
-- active center differencial 

local M = {}
M.type = "auxilliary"
M.relevantDevice = "differential"

local pow = math.pow
local abs = math.abs
local min = math.min
local max = math.max
local sqrt = math.sqrt
local cos = math.cos

local wheels = {} --cache for holding all our wheels
local otherWheelOnAxle = {}

local frontLeftWheelName = ""
local frontRightWheelName = ""
local rearRightWheelName = ""
local rearLeftWheelName = ""

local desiredYawRateSteering = 0
local desiredYawRateAcceleration = 0
local yawRate = 0
local speed = 0 --m/s
local wheelAngleFront = 0
local wheelAngleRear = 0
local desiredYawRate = 0
local yawDifference = 0

local squaredCharacteristicSpeed = 0

local wheelBase = 0 --m
local distanceCOGFrontAxle = 0 --m
local distanceCOGRearAxle = 0 --m
local trackWidth = 0
local mass = 0 --kg

local rearPowerBias = 0
local overSteerResponse = 0.4
local counterOverSteerResponse = 0.4
local underSteerResponse = 0.4
local counterUnderSteerResponse = 0.4

local transfercase = nil
local diffMode = 0
local diffLockPow = 0.05
local oldDiffLockCoef = 0
local newDiffLockCoef = 0
local diffLockSmooth = 0.5

local function updateWheelsIntermediate()
  if not transfercase then return end
  
  local throttle = electrics.values['throttle_input']
  local brake = electrics.values['brake_input']
  local steer = electrics.values['steering_input']
  local absSteer = abs(steer)
  local lateralG = sensors.gx2
  local absLateralG = abs(lateralG)
  local speed = electrics.values.wheelspeed
  
  
  if electrics.values['diffMode'] == 0 then  --Tarmac
	  rearPowerBias = 0.65
	  overSteerResponse = 0.4
	  counterOverSteerResponse = 0.2
	  underSteerResponse = 0.4
	  counterUndSteerResponse = 0.4
  end 
  
  if electrics.values['diffMode'] == 1 then  --Dirt
	  rearPowerBias = 0.6
	  overSteerResponse = 0.4
	  counterOverSteerResponse = 0.15
	  underSteerResponse = 0.3
	  counterUndSteerResponse = 0.4
  end
  
  if electrics.values['diffMode'] == 2 then --Snow
	  rearPowerBias = 0.5
	  overSteerResponse = 0.3
	  counterOverSteerResponse = 0.15
	  underSteerResponse = 0.3
	  counterUndSteerResponse = 0.3
  end
  
  if electrics.values['diffMode'] == 3 then --Drift
	  rearPowerBias = 0.90
	  overSteerResponse = 1.0
	  counterOverSteerResponse = 2.0
	  underSteerResponse = 0.1
	  counterUndSteerResponse = 0.1
  end
  
	
	  

	  local vectorForward = obj:getDirectionVector()
	  local vectorUp = obj:getDirectionVectorUp()
	  local vectorRight = vectorForward:cross(vectorUp)
	  local wheelFront = input.steering < 0 and wheels[frontRightWheelName] or wheels[frontLeftWheelName]
	  local wheelRear = input.steering < 0 and wheels[rearRightWheelName] or wheels[rearLeftWheelName]

	  wheelAngleFront = math.acos(obj:nodeVecPlanarCos(wheelFront.node1, wheelFront.node2, vectorRight, vectorForward))
	  wheelAngleRear = math.acos(obj:nodeVecPlanarCos(wheelRear.node1, wheelRear.node2, vectorRight, vectorForward))
	  
		  if wheelAngleFront > 1.5708 then
		wheelAngleFront = (math.pi - wheelAngleFront)
	  end
	  wheelAngleFront = wheelAngleFront * fsign(-input.steering)
	  if wheelAngleFront ~= wheelAngleFront then
		wheelAngleFront = 0
	  end

	  if wheelAngleRear > 1.5708 then
		wheelAngleRear = (math.pi - wheelAngleRear)
	  end
	  wheelAngleRear = wheelAngleRear * fsign(-input.steering)
	  if wheelAngleRear ~= wheelAngleRear then
		wheelAngleRear = 0
	  end
	  
	  yawRate = yawSmooth:get(-obj:getYawAngularVelocity())

	  --calculate expected yaw rate based on steering angle
	  desiredYawRateSteering = ((wheelAngleFront / wheelBase) * (speed / (1 + ((speed * speed / squaredCharacteristicSpeed)))))
	  --calculate expected yaw rate based on Gs
	  desiredYawRateAcceleration = 10 / (speed + 1e-30)

	  --get the resulting desired yaw rate (smallest) and make sure to use the sign from the steering part (acceleration part is always positive)
	  desiredYawRate = fsign(desiredYawRateSteering) * min(abs(desiredYawRateSteering), abs(desiredYawRateAcceleration))
	  desiredYawRate = desiredYawSmooth:get(desiredYawRate)

	  local counterSteerFlag = false
	  if yawRate * desiredYawRate < 0 then --check if we are counter steering while oversteering.
		desiredYawRate = -desiredYawRate --If we do, we need to adjust our desired yaw rate because its sign is wrong at this point (since we are steering in the "wrong" direction
		counterSteerFlag = true --we need to save this information because we need to alter the wheel that needs to be braked
	  end

	  yawDifference = yawRate - desiredYawRate --calculate the difference between expected yaw and actual yaw, ~0 means we're all good, > 0 means oversteer, < 0 means understeer
	  local absYawDifference = abs(yawDifference)

	  
	    local slideLockUp 

		if abs(yawRate) > abs(desiredYawRate) or counterSteerFlag then --Oversteer
		  slideLockUp = -pow(min(absYawDifference, 1.0),overSteerResponse) * rearPowerBias
		  
		  if counterSteerFlag then
			slideLockUp = -pow(min(absYawDifference, 1.0),counterOverSteerResponse) * rearPowerBias
		  end
		  		  
		else --Understeer
		  slideLockUp =  pow(min(absYawDifference, 1.0),underSteerResponse) * (1.0 - rearPowerBias)

			if counterSteerFlag then
			   --switch to the FRONT wheel when braking, otherwise we would work AGAINST the driver while counter steering
			   --also reduce the severity of the brake action for a smoother drift
			   slideLockUp = -pow(min(absYawDifference, 1.0),counterUnderSteerResponse) * rearPowerBias
			end
		  
		end
	  
	  
	  --local slideLockUp = pow(min(abs(lateralG/11 + steer), 0.7), 1.0)
	  
	  
	  local wheelSpeeds = {}
	  
	  local frontAxleAveSpeed = 0
	  local rearAxleAveSpeed = 0
	  
	  for _,w in pairs(powertrain.wheels) do
		local wheelSpeed = w.speedSmoother:get(w.angularVelocity * w.wheelDir)
		wheelSpeeds[w.name] = wheelSpeed
		--log('I', "Wheel Name ", w.name)
	  end
	  
	  
	  
	  
	  frontAxleAveSpeed = (wheelSpeeds[frontLeftWheelName] + wheelSpeeds[frontRightWheelName]) * 0.5
	  --frontAxleAveSpeed = min(wheelSpeeds[frontLeftWheelName], wheelSpeeds[frontRightWheelName])
	  rearAxleAveSpeed = (wheelSpeeds[rearLeftWheelName] + wheelSpeeds[rearRightWheelName]) * 0.5
	  
	  local axleSpeedRatio = rearAxleAveSpeed / frontAxleAveSpeed
	  
	  local frontPowerBias = min((max((axleSpeedRatio - 1.05),0) * 10.0), 1)
	  transfercase.lsdLockCoef = min((frontPowerBias * 0.9),1)
	  transfercase.diffTorqueSplitA = slideLockUp + rearPowerBias
	  transfercase.diffTorqueSplitB = 1 - transfercase.diffTorqueSplitA
  
  if input.parkingbrake > 0.5 then
    transfercase.lsdLockCoef = 0
  end  
    
	transfercase.lsdRevLockCoef = transfercase.lsdLockCoef
	
  local split = string.format("%.2f", transfercase.diffTorqueSplitA)
  local lock = string.format("%.2f", transfercase.lsdLockCoef) 
  local output1 = string.format("%.2f", transfercase.outputAV1)
  local slide = string.format("%.2f", lateralG/9.8 + steer)
  local yawDiff = string.format("%.2f", yawDifference)
  --log('I', "Yaw Differences ", yawDiff)
  log('I', "Torque Split ", split)
  --log('I', "Diff Lock ", lock)
  --log('I', "Diff Output1 ", output1)
  --log('I', "Slide ", slide)
  
  
end

local function calculateCharacteristicSpeed()
  local eg = (mass * (109000 * distanceCOGRearAxle - 173000 * distanceCOGFrontAxle)) / (173000 * 109000 * wheelBase)
  local characteristicSpeed = sqrt(wheelBase / abs(eg + 1e-30)) --guard against infinity
  
  return characteristicSpeed
end

local function calculateAxleDistances()

  

  wheelBase = obj:nodeLength(wheels[frontRightWheelName].node1, wheels[rearRightWheelName].node1) --calculate wheelbase from the distance of the front and rear wheels

  local tmp = vec3(0,0,0)
  local totalMass = 0
  for _,v in pairs(v.data.nodes) do
    tmp = tmp + vec3(v.pos) * v.nodeWeight
    totalMass = totalMass + v.nodeWeight
  end

  local realCOG = tmp / totalMass

  --Find the positions of the front and rear axle
  local frontAxlePos = 0
  local rearAxlePos = 0
  local twLeft = 0
  local twRight = 0
  for _,n in pairs(v.data.nodes) do
    if n.cid == wheels[frontRightWheelName].node1 then
      frontAxlePos = n.pos.y
      twRight = n.pos.x
    elseif n.cid == wheels[rearRightWheelName].node1 then
      rearAxlePos = n.pos.y
    elseif n.cid == wheels[frontLeftWheelName].node1 then
      twLeft = n.pos.x
    end
  end

  distanceCOGFrontAxle = abs(realCOG.y - frontAxlePos)
  distanceCOGRearAxle = abs(realCOG.y - rearAxlePos)

  trackWidth = abs(twLeft - twRight)

  
end

local function init(jbeamData)
  transfercase = powertrain.getDevice(jbeamData.transfercaseName)
end


local function initSecondStage()
  
  if transfercase then
    transfercase.lsdLockCoef = 0.10
	transfercase.lsdRevLockCoef = 0.10
	transfercase.diffTorqueSplit = 0.98
	electrics.values['diffMode'] = 0
  end
  
  local avgWheelPos = vec3(0,0,0)
  local wheelCount = 0
  wheels = {}
  
  for _,wd in pairs(powertrain.wheels) do
    local name = wd.name
    wheels[name] = wd
    wheels[name].speedSmoother =  newExponentialSmoothing(500)
    --log('I', "Wheel Name ", name)
  end
  
  
  for _,w in pairs(powertrain.wheels) do
    
    local wheelNodePos = v.data.nodes[wheels[w.name].node1].pos --find the wheel position
    avgWheelPos = avgWheelPos + wheelNodePos --sum up all positions
    wheelCount = wheelCount + 1
  end
  
  local vectorForward = vec3(v.data.nodes[v.data.refNodes[0].ref].pos) - vec3(v.data.nodes[v.data.refNodes[0].back].pos) -- vec3(obj:getDirectionVector()) --vector facing forward
  local vectorUp = vec3(v.data.nodes[v.data.refNodes[0].up].pos) - vec3(v.data.nodes[v.data.refNodes[0].ref].pos)

  local vectorRight = vectorForward:cross(vectorUp) --vector facing to the right

  --iterate over all wheels 
  for _,w in pairs(powertrain.wheels) do
    local wheelNodePos = vec3(v.data.nodes[wheels[w.name].node1].pos) --find the wheel position
    local wheelVector = wheelNodePos - avgWheelPos --create a vector from our "center" to the wheel
    local dotForward = vectorForward:dot(wheelVector) --calculate dot product of said vector and forward vector
    local dotLeft = vectorRight:dot(wheelVector) --calculate dot product of said vector and left vector

    if dotForward >= 0 then
      if dotLeft >= 0 then
        
		frontRightWheelName = w.name --this case can only mean it's our front right wheel
      else
        frontLeftWheelName = w.name -- ...
      end
    else
      if dotLeft >= 0 then
        rearRightWheelName = w.name-- ...
      else
        rearLeftWheelName = w.name-- ...
      end
    end
  end
  
  --frontRightWheelName = "FR"
  --frontLeftWheelName = "FL"
  --rearRightWheelName = "RR"
  --rearLeftWheelName = "RL"
  
  yawSmooth = newExponentialSmoothing(5)
  desiredYawSmooth = newExponentialSmoothing(20)
  
  calculateAxleDistances()
  local statsObj = obj:calcBeamStats()
  mass = statsObj.total_weight --simply the mass of the car
  
  local characteristicSpeed = calculateCharacteristicSpeed()
  squaredCharacteristicSpeed = characteristicSpeed * characteristicSpeed
  
  M.updateWheelsIntermediate = updateWheelsIntermediate
  M.updateGFX = updateGFX
  
end



M.init = init
M.initSecondStage   = initSecondStage
--M.reset = init
M.updateWheelsIntermediate = nil
M.updateGFX = nil
--M.updateGFX = updateGFX


return M